package net.ko.utils;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import net.ko.bean.KClass;
import net.ko.bean.KListClass;
import net.ko.kobject.KObject;
import net.ko.persistence.annotation.KProcessAnnotation;

public class KPackageUtils {
	public static KListClass getKClasses(String packageName) throws ClassNotFoundException, IOException {
		return getKClasses(packageName, false);
	}

	@SuppressWarnings("rawtypes")
	public static KListClass getKClasses(String packageName, boolean recursive) throws ClassNotFoundException, IOException {
		KListClass kclasses = new KListClass();
		ArrayList<Class> classes = getClasses(packageName, recursive);
		if (classes != null) {
			for (Class cls : classes)
				kclasses.add(new KClass(cls));
		}
		return kclasses;
	}

	@SuppressWarnings("rawtypes")
	public static ArrayList<Class> getClasses(String packageName) throws ClassNotFoundException, IOException {
		return getClasses(packageName, true);
	}

	/**
	 * Scans all classes accessible from the context class loader which belong
	 * to the given package and subpackages.
	 * 
	 * @param packageName
	 *            The base package
	 * @return The classes
	 * @throws ClassNotFoundException
	 * @throws IOException
	 */
	@SuppressWarnings("rawtypes")
	public static ArrayList<Class> getClasses(String packageName, boolean recursive) throws ClassNotFoundException, IOException {
		ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
		assert classLoader != null;
		String path = packageName.replace('.', '/');
		Enumeration<URL> resources = classLoader.getResources(path);
		List<File> dirs = new ArrayList<File>();
		while (resources.hasMoreElements()) {
			URL resource = resources.nextElement();
			dirs.add(new File(KString.urlDecode(resource.getFile())));
		}
		ArrayList<Class> classes = new ArrayList<Class>();
		for (File directory : dirs) {
			classes.addAll(findClasses(directory, packageName, recursive));
		}
		return classes;
	}

	/**
	 * Recursive method used to find all classes in a given directory and
	 * subdirs.
	 * 
	 * @param directory
	 *            The base directory
	 * @param packageName
	 *            The package name for classes found inside the base directory
	 * @return The classes
	 * @throws ClassNotFoundException
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public static List<Class> findClasses(File directory, String packageName, boolean recursive) throws ClassNotFoundException {
		List<Class> classes = new ArrayList<Class>();
		if (!directory.exists()) {
			return classes;
		}
		File[] files = directory.listFiles();
		for (File file : files) {
			if (file.isDirectory() && recursive) {
				assert !file.getName().contains(".");
				classes.addAll(findClasses(file, packageName + "." + file.getName(), recursive));
			} else if (file.getName().endsWith(".class")) {
				Class clazz = Class.forName(packageName + '.' + file.getName().substring(0, file.getName().length() - 6));
				if (KObject.class.isAssignableFrom(clazz)) {
					if (KProcessAnnotation.isEntity((Class<? extends KObject>) clazz))
						classes.add(clazz);
				}
			}
		}
		return classes;
	}
}
